/* compress_db.c */

#include <Pilot.h>
#include <DataPrv.h>

#define __PALMOS__
#include "zlib.h"

#include "pz.h"


/*
 * constants
 */

#define BUFFER_SIZE 2048


/*
 * types
 */

struct out_stream {
    struct stream stream;

    z_stream z;
    VoidPtr buffer;

    ULong total_size;

    progress_callback_t progress_callback;
};


/*
 * out_stream functions
 */

static int open_stream(
	struct out_stream *out, DmOpenRef db, UInt index,
	ULong max_size, progress_callback_t callback)
{
    int err;

    out->buffer = MemPtrNew(BUFFER_SIZE);
    if (out->buffer == NULL) {
	return MEMORY_ERROR;
    }

    out->z.zalloc = NULL;
    out->z.zfree = NULL;
    out->z.opaque = NULL;
    err = deflateInit(&out->z, Z_DEFAULT_COMPRESSION);
    if (err != Z_OK) {
	return ZLIB_ERROR;
    }

    err = open_out_stream(&out->stream, db, index, max_size);
    out->progress_callback = callback;

    return noErr;
}

static int close_stream(struct out_stream *out)
{
    int err;

    err = deflateEnd(&out->z);

    MemPtrFree(out->buffer);
    close_out_stream(&out->stream, out->z.total_in);

    return (err == Z_OK) ? noErr : ZLIB_ERROR;
}


/*
 * deflate functions
 */

static int deflate_block(struct out_stream *out, const void *ptr, ULong size)
{
    ULong max_in;
    int err;

    out->z.next_in = (void *) ptr;
    max_in = out->z.total_in + size;
    do {
	out->z.avail_in = max_in - out->z.total_in;
	err = deflate(&out->z, Z_NO_FLUSH);
	if (err != Z_OK) {
	    return ZLIB_ERROR;
	}
	if (out->progress_callback != NULL
	 && (*out->progress_callback)(out->z.total_in, out->total_size) != 0) {
	    return CANCELED;
	}
	if (out->z.avail_out == 0) {
	    err = write_out_stream(&out->stream, out->buffer, BUFFER_SIZE);
	    if (err != noErr) { return err; }
	    out->z.next_out = out->buffer;
	    out->z.avail_out = BUFFER_SIZE;
	}
    } while (out->z.total_in < max_in);

    return noErr;
}

static int deflate_handle(struct out_stream *out, VoidHand h)
{
    VoidPtr ptr;
    ULong size;
    int err;

    ptr = MemHandleLock(h);
    size = MemHandleSize(h);
    err = deflate_block(out, ptr, size);
    MemHandleUnlock(h);

    return err;
}

static int deflate_finish(struct out_stream *out)
{
    int err;

    while (true) {
	err = deflate(&out->z, Z_FINISH);
	if (err == Z_STREAM_END) {
	    break;
	} else if (err == Z_OK && out->z.avail_out == 0) {
	    err = write_out_stream(&out->stream, out->buffer, BUFFER_SIZE);
	    if (err != noErr) { return err; }
	    out->z.next_out = out->buffer;
	    out->z.avail_out = BUFFER_SIZE;
	} else {
	    return ZLIB_ERROR;
	}
    }

    if (out->z.avail_out < BUFFER_SIZE) {
	err = write_out_stream(
		&out->stream, out->buffer, BUFFER_SIZE - out->z.avail_out);
	if (err != noErr) { return err; }
    }

    return noErr;
}


/*
 * compress records
 */

static int deflate_records(
	struct out_stream *out, DmOpenRef db, DatabaseHdrPtr db_header,
	UInt card_no, LocalID app_info, LocalID sort_info)
{
    RecordEntryType entry;
    UInt record_attrs;
    ULong unique_id;
    VoidHand h;
    int i, err;

    db_header->recordList.numRecords = DmNumRecords(db);
    out->total_size = sizeof *db_header
	    + sizeof entry * db_header->recordList.numRecords;

    if (app_info != NULL
     && (h = (VoidHand) MemLocalIDToGlobal(app_info, card_no)) != NULL) {
	db_header->appInfoID = out->total_size;
	out->total_size += MemHandleSize(h);
    }

    if (sort_info != NULL
     && (h = (VoidHand) MemLocalIDToGlobal(sort_info, card_no)) != NULL) {
	db_header->sortInfoID = out->total_size;
	out->total_size += MemHandleSize(h);
    }

    err = deflate_block(out, db_header, sizeof *db_header - sizeof(Word));
    if (err != noErr) { return err; }

    for (i = 0; i < db_header->recordList.numRecords; ++i) {
	err = DmRecordInfo(db, i, &record_attrs, &unique_id, NULL);
	if (err != noErr) {
	    return DB_ERROR;
	}
	entry.localChunkID = out->total_size;
	entry.attributes = record_attrs;
	entry.uniqueID[0] = (unique_id >> 16) & 0xff;
	entry.uniqueID[1] = (unique_id >> 8) & 0xff;
	entry.uniqueID[2] = unique_id & 0xff;

	err = deflate_block(out, &entry, sizeof entry);
	if (err != noErr) { return err; }

	h = DmQueryRecord(db, i);
	if (h != NULL) {
	    out->total_size += MemHandleSize(h);
	}
    }

    err = deflate_block(out, &db_header->recordList.firstEntry, sizeof(Word));
    if (err != noErr) { return err; }

    if (app_info != NULL
     && (h = (VoidHand) MemLocalIDToGlobal(app_info, card_no)) != NULL) {
	err = deflate_handle(out, h);
	if (err != noErr) { return err; }
    }

    if (sort_info != NULL
     && (h = (VoidHand) MemLocalIDToGlobal(sort_info, card_no)) != NULL) {
	err = deflate_handle(out, h);
	if (err != noErr) { return err; }
    }

    for (i = 0; i < db_header->recordList.numRecords; ++i) {
	h = DmQueryRecord(db, i);
	if (h != NULL) {
	    err = deflate_handle(out, h);
	    if (err != noErr) { return err; }
	}
    }

    return NULL;
}


/*
 * compress resources
 */

static int deflate_resources(
	struct out_stream *out, DmOpenRef db, DatabaseHdrPtr db_header)
{
    RsrcEntryType entry;
    VoidHand h;
    int i, err;

    db_header->recordList.numRecords = DmNumResources(db);
    out->total_size = sizeof *db_header
	    + sizeof entry * db_header->recordList.numRecords;

    err = deflate_block(out, db_header, sizeof *db_header - sizeof(Word));
    if (err != noErr) { return err; }

    for (i = 0; i < db_header->recordList.numRecords; ++i) {
	err = DmResourceInfo(db, i, &entry.type, &entry.id, NULL);
	if (err != noErr) {
	    return DB_ERROR;
	}
	entry.localChunkID = out->total_size;

	err = deflate_block(out, &entry, sizeof entry);
	if (err != noErr) { return err; }

	h = DmGetResourceIndex(db, i);
	if (h != NULL) {
	    out->total_size += MemHandleSize(h);
	    DmReleaseResource(h);
	}
    }

    err = deflate_block(out, &db_header->recordList.firstEntry, sizeof(Word));
    if (err != noErr) { return err; }

    for (i = 0; i < db_header->recordList.numRecords; ++i) {
	h = DmGetResourceIndex(db, i);
	if (h != NULL) {
	    err = deflate_handle(out, h);
	    DmReleaseResource(h);
	    if (err != noErr) { return err; }
	}
    }

    return NULL;
}


/*
 * compress database
 */

static int deflate_db(struct out_stream *out, UInt card_no, LocalID db_id)
{
    DatabaseHdrType db_header;
    LocalID app_info, sort_info;
    DmOpenRef db;
    Boolean data_db;
    char *p;
    int n, err;

    MemSet(&db_header, sizeof db_header, 0);
    err = DmDatabaseInfo(card_no, db_id,
	    db_header.name, &db_header.attributes, &db_header.version,
	    &db_header.creationDate, &db_header.modificationDate,
	    &db_header.lastBackupDate, &db_header.modificationNumber,
	    &app_info, &sort_info, &db_header.type, &db_header.creator);
    if (err != noErr) {
	return DB_ERROR;
    }

    db_header.attributes &= ~ (dmHdrAttrOpen | dmHdrAttrReadOnly);
    data_db = (db_header.attributes & dmHdrAttrResDB) == 0;

    p = (char *) out->buffer + sizeof(struct pz_header);
    StrCopy(p, db_header.name);
    StrCat(p, data_db ? DATA_DB_EXT : RES_DB_EXT);

    n = sizeof(struct pz_header) + StrLen(p) + 1;
    out->z.next_out = (char *) out->buffer + n;
    out->z.avail_out = BUFFER_SIZE - n;

    db = DmOpenDatabase(card_no, db_id, dmModeReadOnly);
    if (db == NULL) {
	return DB_ERROR;
    }

    if (data_db) {
	err = deflate_records(
		out, db, &db_header, card_no, app_info, sort_info);
    } else {
	err = deflate_resources(out, db, &db_header);
    }

    DmCloseDatabase(db);

    if (err != noErr) { return err; }

    err = deflate_finish(out);
    if (err != noErr) { return err; }

    if (out->z.total_in != out->total_size) {
	return DATA_ERROR;
    }

    return noErr;
}

int compress_db(DmOpenRef arch_db, UInt index, UInt card_no, LocalID db_id,
	ULong *uncompr_size, ULong *compr_size, progress_callback_t callback)
{
    struct out_stream out;
    ULong db_size;
    int err;

    err = DmDatabaseSize(card_no, db_id, NULL, &db_size, NULL);
    if (err != noErr) {
	return DB_ERROR;
    }

    db_size += (db_size + 999) / 1000 + 12;
    err = open_stream(&out, arch_db, index, db_size, callback);
    if (err != noErr) { return err; }

    ((struct pz_header *) out.buffer)->id1 = PZ_ID1;
    ((struct pz_header *) out.buffer)->id2 = PZ_ID2;
    ((struct pz_header *) out.buffer)->cm = PZ_CM_DEFLATED;
    ((struct pz_header *) out.buffer)->flg = PZ_FLG_FNAME;
    ((struct pz_header *) out.buffer)->mtime = 0;
    ((struct pz_header *) out.buffer)->xfl = 0;
    ((struct pz_header *) out.buffer)->os = PZ_OS_UNKNOWN;

    err = deflate_db(&out, card_no, db_id);

    if (err == noErr) {
	err = close_stream(&out);
    } else {
	close_stream(&out);
	remove_stream(arch_db, out.stream.first_index, out.stream.num_records);
    }

    if (err == noErr) {
	*uncompr_size = out.total_size;
	*compr_size = out.stream.cur_start;
    }

    return err;
}
